static char rcsid[] = "$Id: source.c,v 1.5 1992/03/25 17:48:26 jtsillas Exp $";

/*****************************************************************************
 *
 *  Copyright 1989 The University of Texas at Austin
 *  Copyright 1990 Microelectronics and Computer Technology Corporation
 *  Copyright 1990 Thomson Consumer Electronics, Inc.
 *  Copyright 1991 Bull HN Worldwide Info Systems, Inc.
 *  Copyright 1993 Philippe-Andre Prindeville, Telecom Paris.
 *
 *****************************************************************************/

/*  source.c
 *
 *    Create the source window and handle display of file.
 *
 *    source_init(): 	Initialization routine.
 *    SelectWord():     Replaces the standard select translation (static).
 *    CreateSourceWindow(): Create the source window.
 *    BuildLinePos():	Build an array of starting text position of each line
 *                      (static).
 *    LookUpFileTable(): Check out source file info from a file table (static).
 *    SaveDisplayedFileInfo(): Records displayed file info into file table
 *                      (static).
 *    DisplayFile():	Display a file on the source window (static).
 *    LoadFile():	Search for a file and open it for display.
 *    LoadCurrentFile(): Load the CurrentFile.
 *    CreateCommandButtons(): Create the command buttons (static).
 *    PopupCommands():  Manage the commands dialog.
 *    MakeDirList():    Create a list of directories to search for source.
 *    GetPathname():    Get a path for a file using the dirList.
 */

#include <X11/Xos.h>
#include <sys/stat.h>
#ifdef HAVE_MMAP
#include <sys/mman.h>
#endif
#include <pwd.h>
#include "global.h"
#include "bitmaps.h"
#if defined(SCO)
#   include <fcntl.h>
#endif

#include <Xm/Form.h>
#include <Xm/Text.h>
#include <Xm/ScrolledW.h>
#include <Xm/ScrollBar.h>
#include <Xm/ArrowB.h>
#include <Xm/PushB.h>
#include <Xm/DrawingA.h>

#define	MAXDIRS	256			/* max number of dirs in dirList */

static void SelectWord(Widget, XEvent*, String*, Cardinal*);
static void resetLineBounds(Widget, XEvent*, String*, Cardinal*);
static void CreateCommandButtons(Widget);
static void DisplayFile(FileRec*);
static void SaveDisplayedFileInfo(void);
static int LookUpFileTable(char*, char*, FileRec**);
static void BuildLinePos(FileRec*);

enum { PAGE_LEFT, PAGE_RIGHT };

char		CurrentFile[MAXNAME];	/* current contents of file variable */
Widget		sourceWindow,		/* text window for source file */
                sourceDrawingArea,
                sourceScroll,
                scrollV,
                sourceForm,
                commandDialog;

FileRec  	*displayedFile;		/* pointer to table entry of currently
					   displayed file */

static FileRec	**fileTable;		/* table of file records */
static int	fileTableSize;		/* size of file table */
static char 	*dirList[MAXDIRS];	/* list of dirs for searching files */
static short    maxcol;
static short    currvalue;

void source_init()
{
    dirList[0] = NULL;
}

/*  This is my own select word routine to replace the standard action
 *  procedure provided by the Text widget.
 *  It selects a word delimited by DELIMITERS, not whitespace.
 */
static void SelectWord(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal *num_params;
{
  XmTextPosition left, right, center, last;
  char *valuestring;
  
  XtCallActionProc(w, "set-insertion-point", event, params, *num_params);
  center = XmTextGetInsertionPosition(w);
  valuestring = XmTextGetString(w);
  last = XmTextGetLastPosition(w);
  
  if(strchr(DELIMITERS, valuestring[center])) return;
  left = center;
  right = center;
  
  while(left>0 && !strchr(DELIMITERS, valuestring[left-1])) left--;
  while(right<last && !strchr(DELIMITERS, valuestring[right])) right++;
  
  XmTextSetSelection(w, left, right, 0);
  XtFree(valuestring);
}

/*
 * this function updates the ScrollBar value, as well as potentially
 * forcing an activation callback.
 */

void ScrollBarSetValues(wid, value, sliderSize, increment, pageIncrement, activate)
    Widget wid;
    int value, sliderSize, increment, pageIncrement;
    Boolean activate;
{
    int dummy1, dummy2, dummy3, dummy4;

    /* not really "lite", but it works... */
    XmScrollBarGetValues(wid, &dummy1, &dummy2, &dummy3, &dummy4);
    if (value == -1) value = dummy1;
    if (sliderSize == -1) sliderSize = dummy2;
    if (increment == -1) increment = dummy3;
    if (pageIncrement == -1) pageIncrement = dummy4;
    XmScrollBarSetValues(wid, value, sliderSize, increment, pageIncrement, activate);
}

/* If the scrollbar is moved by dragging or clicking, we must update
 * the topCharacter of the text widget to reflect the new "home"...
 */
static void
ScrollBarCB(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     XmScrollBarCallbackStruct *call_data;
{
	FileRec *file = displayedFile;

	if (file) {
		file->topline = call_data->value;
		file->bottomline = MIN(file->topline + file->lines - 1,
				       file->lastline);
		XmTextSetTopCharacter(sourceWindow,
				      file->linepos[file->topline]);
		UpdateSigns(file);
	}
}

/* allow flipping page-right or page-left in the text window */
static void
FlipTextCB(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     XmPushButtonCallbackStruct *call_data;
{
     switch ((int) client_data) {
     case PAGE_LEFT:
     case PAGE_RIGHT:
	XtCallActionProc(sourceWindow,
			 ((int) client_data == PAGE_RIGHT) ? "page-right" : "page-left" ,
			 call_data->event, NULL, 0);
      break;
    default:
	fprintf(stderr, "FlipTextCB: Can't happen\n");
    }
}

/* if the contents of the file change, we must set the maximum, the
 * current value, etc.  We also set the (page) increments and the
 * slider size for good measure.  Note:  displayedFile *must* point
 * to the right file at this time.
 */
static void ContentsChangedCB(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
     short rows;

     XtGetValue(sourceWindow, XmNrows, &rows);

       /* If the window is bigger than the total length of the file,
	* unmanage the scrollbar.
	*/
     if (rows >= displayedFile->lastline)
       XtUnmanageChild(scrollV);
     else {
       XtVaSetValues(scrollV,
		     XmNminimum, 1,
		     XmNvalue, displayedFile->topline,
		     XmNsliderSize, rows,
		     XmNmaximum, displayedFile->lastline + 1,
		     NULL);
       ScrollBarSetValues(scrollV, -1, -1, 1, rows - 1, False);

       XtManageChild(scrollV);
     }
}

static void ResizeSourceWindow(w, closure, event, cont)
     Widget w;
     XtPointer closure;
     XEvent *event;
     Boolean *cont;
{
    short rows;

    switch (event->type) {
    case ConfigureNotify:
	XtGetValue(w, XmNrows, &rows);

	/* note: when we resize, we try to keep the same home ("top")
	 * position.  this means that the current line may conceivably
	 * become hidden.  since resizing doesn't happen often, it is
	 * simpler just to let the user scroll down to the new point.
	 */
        if (displayedFile) {
	    register FileRec *file = displayedFile;

	    file->lines = rows;
	    file->bottomline = MIN(file->topline + file->lines - 1,
				   file->lastline);
	    /* This callback appears to do the right thing. */
	    ContentsChangedCB(NULL, NULL, NULL);
	    UpdateSigns(file);
    }
	break;
     default:
#if 0
fprintf(stderr, "ResizeSourceWindow: event type %d\n", event->type);
#endif
        break;
     }
}

#define BUTTONSNUM 18		/* Number of buttons */
#define NCOLUMNS 2		/* Should divide BUTTONSNUM evenly */
#define SPACEFACTOR 32		/* These number can be diddled. */
#define BUTTONWIDTH 32
#define FRACTIONBASE 	     ((BUTTONSNUM * SPACEFACTOR) / NCOLUMNS) 

static void CreateCommandButtons(parent)
     Widget parent;
{
  static buttonstruct commandbuttons[] = {
    { NULL, "stackbutton", stack_bits, stack_width, stack_height, 
	&app_resources.button_color, DoIt, "info stack\n" },
    { NULL, "runbutton", run_bits, run_width, run_height, 
	&app_resources.button_color, DoIt, "run\n" },
    { NULL, "upbutton", up_bits, up_width, up_height, 
	&app_resources.updown_color, DoIt, "up\n" },
    { NULL, "localsbutton", locals_bits, locals_width, locals_height, 
	&app_resources.button_color, DoIt, "info locals\n" },
    { NULL, "downbutton", down_bits, down_width, down_height, 
	&app_resources.updown_color, DoIt, "down\n" },
    { NULL, "argsbutton", args_bits, args_width, args_height, 
	&app_resources.button_color, DoIt, "info args\n" },
    { NULL, "printbutton", print_bits, print_width, print_height, 
	&app_resources.button_color, Print, NULL },
    { NULL, "printsbutton", prints_bits, prints_width, prints_height, 
	&app_resources.button_color, Print, (void *) 1 },
    { NULL, "nextbutton", next_bits, next_width, next_height, 
	&app_resources.button_color, DoIt, "next\n" },
    { NULL, "contbutton", cont_bits, cont_width, cont_height, 
	&app_resources.button_color, DoIt, "cont\n" },
    { NULL, "stepbutton", step_bits, step_width, step_height, 
	&app_resources.button_color, DoIt, "step\n" },
    { NULL, "finishbutton", finish_bits, finish_width, finish_height, 
	&app_resources.button_color, DoIt, "finish\n" },
    { NULL, "breakbutton", stop_bits, stop_width, stop_height, 
	&app_resources.stop_color, Break, "break" },
    { NULL, "clearbutton", clear_bits, clear_width, clear_height, 
	&app_resources.stop_color, Clear, NULL },
    { NULL, "dispbutton", disp_bits, disp_width, disp_height, 
	&app_resources.button_color, Display_, NULL },
    { NULL, "undispbutton", undisp_bits, undisp_width, undisp_height, 
	&app_resources.button_color, Undisplay, NULL },
    { NULL, "pageRbutton", pageR_bits, pageR_width, pageR_height, 
	&app_resources.arrow_color, FlipTextCB, (XtPointer) PAGE_RIGHT},
    { NULL, "pageLbutton", pageL_bits, pageL_width, pageL_height, 
	&app_resources.arrow_color, FlipTextCB, (XtPointer) PAGE_LEFT},
  };
  
  Pixel bg;
  int screen;
  Display *topdisplay;
  int window1;
  int depth1;
  int buttonindex;
  XtCallbackList buttoncalls;

  XtGetValue(sourceForm, XmNbackground, &bg);
  topdisplay = XtDisplay(toplevel);
  screen = DefaultScreen(topdisplay);
  window1 = DefaultRootWindow(topdisplay);
  depth1 = DefaultDepth(topdisplay, screen);
  
  for (buttonindex = 0; buttonindex < BUTTONSNUM; buttonindex++)
    {
      commandbuttons[buttonindex].button = 
	XtVaCreateManagedWidget(commandbuttons[buttonindex].buttonname,
			      xmPushButtonWidgetClass, parent,
    			      XmNbottomAttachment, XmATTACH_POSITION,
    			      XmNtopAttachment, XmATTACH_POSITION,
    			      XmNrightAttachment, XmATTACH_FORM,
    			      XmNwidth, BUTTONWIDTH,
    			      XmNlabelType, XmPIXMAP,
    			      XmNhighlightOnEnter, True,
      			      XmNlabelPixmap, XCreatePixmapFromBitmapData(
				    topdisplay, window1,
				    commandbuttons[buttonindex].labelbits,
				    commandbuttons[buttonindex].labelwidth,
				    commandbuttons[buttonindex].labelheight, 
				    *(commandbuttons[buttonindex].label_color),
				    bg, depth1),
      			       XmNrightOffset,
	       			(buttonindex%NCOLUMNS)?(BUTTONWIDTH*(buttonindex%NCOLUMNS)):0,
      			       XmNtopPosition,
				((buttonindex/NCOLUMNS) * SPACEFACTOR),
      			       XmNbottomPosition,
	       			(((buttonindex/NCOLUMNS) + 1) * SPACEFACTOR),
			       NULL);

      XtAddCallback(commandbuttons[buttonindex].button, 
		    XmNactivateCallback, commandbuttons[buttonindex].CB, 
		    commandbuttons[buttonindex].CBarg);
    }
}

/* 
 *  On top of a form widget, we have a text widget, label
 *  widgets for the stop sign, arrow sign, and updown signs.
 */
void CreateSourceWindow(parent)
     Widget parent;
{
  static XtActionsRec command_actions[] = {
    { "SelectWord", (XtActionProc) SelectWord },
    { NULL, NULL }
  };

  sourceScroll = XtVaCreateManagedWidget("sourceScroll", 
				      xmScrolledWindowWidgetClass, parent,
/* important enough not to let the user muddle with it... */
				      XmNscrollingPolicy, XmAPPLICATION_DEFINED,
				      NULL); 

  scrollV = XtVaCreateWidget("scrollV", xmScrollBarWidgetClass, sourceScroll,
			     XmNminimum, 1,
			     XmNmaximum, 2,
			     XmNvalue, 1,
			     XmNsliderSize, 1,
			     NULL);

  sourceForm = XtVaCreateManagedWidget("sourceForm", xmFormWidgetClass,
				     sourceScroll,
				     XmNfractionBase, FRACTIONBASE,
				     NULL);
  
  XmScrolledWindowSetAreas(sourceScroll, NULL, scrollV, sourceForm);
  
  XtAddCallback(scrollV, XmNdragCallback, (XtCallbackProc) ScrollBarCB, NULL);
  XtAddCallback(scrollV, XmNvalueChangedCallback, (XtCallbackProc) ScrollBarCB, NULL);
  
  sourceDrawingArea = XtVaCreateManagedWidget("sourceDrawingArea", xmDrawingAreaWidgetClass,
				    sourceForm, NULL);
  
  XtAddCallback(sourceDrawingArea, XmNexposeCallback, (XtCallbackProc) DrawSigns, NULL);
  XtAddCallback(sourceDrawingArea, XmNinputCallback, (XtCallbackProc) ClickOnDrawingAreaCB, NULL);

  if(app_resources.useCommandDialog)
      sourceWindow = XtCreateManagedWidget("sourceWindow", xmTextWidgetClass,
					 sourceForm, NULL, 0);
  else
    sourceWindow = XtVaCreateManagedWidget("sourceWindow", xmTextWidgetClass,
					   sourceForm,
      					   XmNrightOffset, BUTTONWIDTH * NCOLUMNS,
					   NULL);
  XtVaSetValues(sourceWindow,
		XmNeditable, False,
		XmNleftWidget, sourceDrawingArea,
		NULL);
    
  XtAddCallback(sourceWindow, XmNvalueChangedCallback,
		(XtCallbackProc) ContentsChangedCB, NULL);

  /* would be nice if the window had a bloody resize callback list! */
  XtAddEventHandler(sourceWindow, StructureNotifyMask, False,
		    (XtEventHandler) ResizeSourceWindow, NULL);
 
  XtAppAddActions(app_context, command_actions, XtNumber(command_actions));

  if(app_resources.useCommandDialog)
    {
      commandDialog = XmCreateFormDialog(parent, "commandDialog",
					 NULL, 0);
      XtSetValue(commandDialog, XmNfractionBase, FRACTIONBASE);
      CreateCommandButtons(commandDialog);
    }
  else
    CreateCommandButtons(sourceForm);
}

/* 
 * Popup the commandDialog Box if we are using it.
 */
void PopupCommands(w, client_data, call_data)
     Widget w;
     XtPointer client_data;
     XmPushButtonCallbackStruct *call_data;
{
  if(app_resources.useCommandDialog)
    XtManageChild(commandDialog);
  else
    {
      UpdateMessageWindow("Not using command button dialog", NULL);
      bell(0);
    }
}

/*
 *  Build the array which gives the starting text position of each line.
 *  > Estimate the number of lines in the file and allocate memory buffer.
 *  > Starting position of line #1 is 0, and is stored in linepos[1].
 *  > Search for '\n' till end of buffer.
 */
static void BuildLinePos(file)
FileRec *file;
{
    char *p;
    int	 line, nlines;

    nlines = MAX(2, file->filesize/CHARS_PER_LINE);
    file->linepos = (XmTextPosition *)
		    XtMalloc ((nlines+1) * sizeof(XmTextPosition));
    p = file->buf;

    file->linepos[0] = 0;
    file->linepos[1] = 0;
    line = 2;
   
    while (*p) {
	if (*p++ == '\n') {
	    if (line == nlines) { 	/* buffer full, need more memory */
                file->linepos = (XmTextPosition *) XtRealloc ((char *) file->linepos, 
			  (nlines + ADD_LINES) * sizeof(XmTextPosition));
		nlines += ADD_LINES;
            }
            file->linepos[line++] = p - file->buf;
	}
    }
    file->lastline = line - 2;
    file->linepos = (XmTextPosition *) XtRealloc       /* shrink to min size */
			((char *) file->linepos, line * sizeof(XmTextPosition));
}


/*
 * Look up the file table for an entry with "filename"
 * If not found, create an entry and initialize proper fields,
 * else, return pointer to entry found.
 */
static int LookUpFileTable(pathname, filename, file)
     char *pathname, *filename;
     FileRec **file;
{
  struct stat fileinfo;
  int  	fd;
  int 	i, j, n;
  
  for (i=0; fileTable && fileTable[i] && i<fileTableSize; i++) {
    if (!strcmp(fileTable[i]->pathname, pathname)) { /* file found */
      if (stat(pathname, &fileinfo) == -1) {
	UpdateMessageWindow("Error: cannot stat file %s", pathname);
	*file = fileTable[i];
	return 0;
      }
      if (fileinfo.st_mtime > fileTable[i]->mtime) { /* file modified */
#ifdef HAVE_MMAP
	munmap(fileTable[i]->buf, fileTable[i]->filesize);
#else
	XtFree((char *)fileTable[i]->buf);
	XtFree((char *)fileTable[i]->linepos);
	XtFree((char *)fileTable[i]);
#endif
	displayedFile = NULL;
	break;
      }
      else
	if (displayedFile && 		/* same as displayed file */
	    strcmp(pathname, displayedFile->pathname) == 0) {
	  *file = NULL;
	  return 0;
	}
	else {
	  *file = fileTable[i];
	  return 0;
	}
    }
  }
  
    /* Record file into file table */
    
  if (i == fileTableSize) {		/* file table full, enlarge it */
    fileTableSize += ADD_SIZE;
    fileTable = (FileRec **) 
      XtRealloc ((char *) fileTable, fileTableSize * sizeof(FileRec *));
    for (j=i; j<fileTableSize; j++)
      fileTable[j] = NULL;
  }
  if ((fd = open(pathname, O_RDONLY)) == -1) {
    UpdateMessageWindow("Error: cannot open file %s", pathname);
    return -1;
  }
  if (fstat(fd, &fileinfo) == -1) {
    UpdateMessageWindow("Error: cannot fstat file %s", pathname);
    close(fd);
    return -1;
  }
  fileTable[i] = (FileRec *) XtMalloc (sizeof(FileRec));
  fileTable[i]->filesize = fileinfo.st_size + 1;
  fileTable[i]->mtime = fileinfo.st_mtime;
#ifdef HAVE_MMAP
  fileTable[i]->buf = mmap(0, fileTable[i]->filesize, PROT_READ,
			   MAP_PRIVATE, fd, 0);
  if (fileTable[i]->buf == (char *)-1)
    {
#else
  fileTable[i]->buf = XtMalloc((int)fileTable[i]->filesize);
  if ((n = read(fd, fileTable[i]->buf,
		(int) fileTable[i]->filesize)) == -1)
    {
      XtFree(fileTable[i]->buf);
      
#endif
      UpdateMessageWindow("Error: cannot read file %s", pathname);
      XtFree((char *) fileTable[i]);
      fileTable[i] = NULL;
      close(fd);
      return -1;
    }
#ifndef HAVE_MMAP
  fileTable[i]->buf[n] = '\0';
#endif
  fileTable[i]->pathname = XtNewString(pathname);
  fileTable[i]->filename = XtNewString(filename);
  fileTable[i]->currentline = 1;
  fileTable[i]->topline = 1;
  fileTable[i]->bottomline = 0;
  BuildLinePos(fileTable[i]);
  close(fd);
  *file = fileTable[i];
  return 0;
}
  
/*  
 *  Remember file position and current line before closing.
 */
  
static void SaveDisplayedFileInfo()
{
  if (displayedFile) {
    displayedFile->topline = TextPositionToLine(
				XmTextGetTopCharacter(sourceWindow));
    displayedFile->currentline = TextPositionToLine(
				     XmTextGetInsertionPosition(sourceWindow));
  }
}

/*   DisplayFile() displays the file onto the source window.  It
 *     uses topline to remember where it was last opened.  But it
 *     must recalculate bottomline because the window size might be
 *     different.
 */
static void DisplayFile(file)
FileRec *file;
{
    short       rows;

    /* don't worry about updating the scrollbar -- the max and min are
     * set automatically, as are the scrolling increments and slider size.
     */
    XtVaSetValues(sourceWindow,
    		  XmNvalue, file->buf,
		  XmNtopCharacter, file->linepos[file->topline],
		  NULL);

    XtGetValue(sourceWindow, XmNrows, &rows);

    file->lines = rows;
    file->bottomline = MIN (file->topline + file->lines - 1, file->lastline);
}

/*  Create a list of directories for searching source files.
 *  It reads the list of directories specified by the user, adding
 *  the current directory into the list if it is not already there.
 */
void MakeDirList(output)
char *output;
{
    char *s, list[LINESIZ], command[LINESIZ];
    int  i;

    for (i=0; dirList[i]; i++)			/* remove old list */
	XtFree(dirList[i]);
    i = 0;

    if (output) {                                        /* create list */
        s = (char *) strtok(output, ":");
        while (s) {
            dirList[i] = XtNewString(s);
            i++;
            s = (char *) strtok(NULL, ":");
        }
        dirList[i] = NULL;
    }
}

/*  Returns the full pathname of a given file.
 *  It searches for the file from a list of directories.
 */
char *GetPathname(filename)
     char *filename;
{
  char pathname[LINESIZ];
  char *normal_file;
  int i;
  
  /* Normalize the file name using normal_file to point to
   * the last element of a path.
   */
  for (normal_file = filename + strlen(filename); normal_file != filename 
       && *normal_file != '/'; normal_file--);
  if (*normal_file == '/') normal_file++;
  if (normal_file == NULL || !strcmp(normal_file, ""))
    return NULL;
  
  for (i=0; dirList[i]; i++) 
    {
      if (!strcmp(dirList[i], "$cwd"))
	sprintf(pathname, "%s/%s", cwd, normal_file);
      else if (!strcmp(dirList[i], "$cdir"))
	sprintf(pathname, "%s/%s", cwd, filename);
      else
	sprintf(pathname, "%s/%s", dirList[i], normal_file);
      if (access(pathname, R_OK) == 0)
	return XtNewString(pathname);
    }

  if (access(filename, R_OK) == 0)
    return XtNewString(filename); /* When all else fails try the obvious. */

  UpdateMessageWindow("File not found: %s", filename);
  return NULL;
}

/*
 * Given a file name, LoadFile attempts to open it and displays it onto
 * the source window:
 *   1. get the full pathname of the file
 *   2. LookUpFileTable() returns a pointer to the file's entry if it's
 *      already in the table; else, creates an entry and return a pointer.
 *   3. save the current displayedFile info
 *   4. display the file
 *   5. update the file label and the various signs on the source window.
 *  LoadFile returns 0 upon successful completion, -1 otherwise.
 */
int LoadFile(filename)
char *filename;
{
    FileRec 	*file;
    char	*pathname;
    
    pathname = GetPathname(filename);
    if (pathname == NULL) { 
	return -1;
    }
    if (LookUpFileTable(pathname, filename, &file) != -1) {
	if (file) {	/* load new file */
	    SaveDisplayedFileInfo();
	    displayedFile = file;
	    DisplayFile(file);
	    UpdateFileLabel(pathname);
	    XmTextClearSelection(sourceWindow,0);
	    XmTextSetInsertionPosition(sourceWindow, 
				       file->linepos[file->currentline]);
	    UpdateLineLabel(file->currentline);
	    UpdateSigns(file);
	}
    	return 0;
    }
    else {		/* LookUpFileTable() fails */
    	return -1;
    }
}

int LoadCurrentFile()
{
    query_gdb("info line\n");
    return LoadFile(CurrentFile);
}

